import asyncio
import time
import traceback
from datetime import datetime
from datetime import timedelta
from typing import List

from asgiref.sync import sync_to_async
import aiohttp
from django.core.management.base import BaseCommand
from django.db import OperationalError
from django.db import transaction
from django.db.models import Q
from pytz import utc

from django.conf import settings
from push_notification.event.models import PushNotificationEvent


class Command(BaseCommand):
    help = "Send events to registered push notification url"

    retry = settings.PUSH_NOTIFICATIONS.get('RETRY', 10)
    retry_interval = settings.PUSH_NOTIFICATIONS.get('RETRY_INTERVAL', 60)
    connection_timeout = settings.PUSH_NOTIFICATIONS.get('CONNECTION_TIMEOUT', 30)
    read_timeout = settings.PUSH_NOTIFICATIONS.get('READ_TIMEOUT', 30)
    connection_per_host = settings.PUSH_NOTIFICATIONS.get('CONNECTIONS_PER_HOST', 10)
    process_butch = settings.PUSH_NOTIFICATIONS.get('PROCESS_BUTCH', 100)

    def handle(self, *args, **kwargs):
        print("push notifications emitter started")

        while True:
            interval = (datetime.utcnow() - timedelta(seconds=self.retry_interval)).replace(tzinfo=utc)
            try:
                with transaction.atomic():
                    events = PushNotificationEvent.actual_objects.select_for_update(nowait=True).filter(
                        Q(retry__isnull=True) | Q(retry__lt=self.retry),
                        Q(last_retried_at__isnull=True) | Q(last_retried_at__lt=interval)
                    ).all()[:self.process_butch]

                    events_len = events.count()
                    if events_len < 1:
                        time.sleep(1)
                        continue

                    print(
                        "proccessing {} events".format(events_len)
                    )

                    asyncio.run(self.emits(events))
            except Exception as e:
                traceback.print_exc()
            time.sleep(1)

    # this needs to be separated like so because transactions do not support async mode in django 4.1
    # https://docs.djangoproject.com/en/4.1/topics/async/#queries-the-orm
    def create_event_tasks(self, session, events):
        tasks = []
        with transaction.atomic():
            for event in events:
                tasks.append(
                    self.emit(session, event)
                )
        return tasks

    async def emits(self, events: List[PushNotificationEvent]):
        conn = aiohttp.TCPConnector(limit_per_host=self.connection_per_host)
        async with aiohttp.ClientSession(
                conn_timeout=self.connection_timeout, read_timeout=self.read_timeout,
                connector=conn) as session:
            tasks = await sync_to_async(self.create_event_tasks)(session, events)
            await asyncio.gather(*tasks)

    async def emit(self, session: aiohttp.ClientSession, event: PushNotificationEvent):
        print(
            "emit event {}".format(event.event_id))

        utcnow = datetime.utcnow().replace(tzinfo=utc)
        event.retry = 1 if not event.retry else event.retry + 1
        event.last_retried_at = utcnow

        try:
            res = await event.post(session)
            res.raise_for_status()
            event.processed_at = utcnow
            print(
                "event {} has been delivered to {}".format(
                    event.event_id, res.url)
            )
        except Exception as e:
            print(
                "error while delivering the event {}, retry {}: {}".format(
                    event.event_id, event.retry, str(e))
            )
        finally:
            await sync_to_async(event.save)()



import fnmatch
import uuid
import aiohttp
from django.db import models
from django.dispatch import receiver
from django.db.models.signals import post_save


class Event(models.Model):
    event_id = models.UUIDField(
        primary_key=True, default=uuid.uuid4, editable=False)
    event_type = models.CharField(max_length=254)
    object_type = models.CharField(max_length=254, null=True, blank=True)
    object_id = models.CharField(max_length=255, null=True, blank=True)
    account_id = models.UUIDField(null=True, blank=True)
    meta_data = models.JSONField(null=True, blank=True)

    created_at = models.DateTimeField(auto_now_add=True)
    registerd_at = models.DateTimeField(null=True)

    class Meta:
        verbose_name = "event"
        verbose_name_plural = "events"

    def data(self) -> dict:
        return {
            "event_id":  str(self.event_id),
            "event_type": self.event_type,
            "object_type": self.object_type if self.object_type else None,
            "object_id": str(self.object_id) if self.object_id else None,
            "account_id": str(self.account_id) if self.account_id else None,
            "meta_data": self.meta_data if self.meta_data else None,
        }

    def register(self):
        push_notifications = PushNotification.actual_objects.filter(
            created_at__lte=self.created_at
        ).all()

        for push_notification in push_notifications:
            if fnmatch.fnmatch(self.event_type, push_notification.pattern):
                PushNotificationEvent.objects.get_or_create(
                    event=self, push_notification=push_notification
                )


@receiver(post_save, sender=Event)
def event_registrator(sender, **kwargs):
    event = kwargs.get("instance", None)
    if event:
        event.register()


class PushNotificationManager(models.Manager):
    def get_queryset(self):
        return super().get_queryset().filter(deleted_at__isnull=True)


class PushNotification(models.Model):
    push_notification_id = models.UUIDField(
        primary_key=True, default=uuid.uuid4, editable=False)

    callback_url = models.URLField(max_length=2000)
    pattern = models.CharField(max_length=10000)

    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)
    deleted_at = models.DateTimeField(null=True)

    objects = models.Manager()
    actual_objects = PushNotificationManager()

    class Meta:
        verbose_name = "push notification"
        verbose_name_plural = "push notifications"
        unique_together = [
            'callback_url', 'pattern',
        ]


class PushNotificationEventManager(models.Manager):
    def get_queryset(self):
        return super().get_queryset().filter(processed_at__isnull=True)


class PushNotificationEvent(models.Model):
    push_notification = models.ForeignKey(
        PushNotification, on_delete=models.CASCADE)
    event = models.ForeignKey('Event', on_delete=models.CASCADE)
    retry = models.IntegerField(null=True, blank=True)

    created_at = models.DateTimeField(auto_now_add=True)
    processed_at = models.DateTimeField(null=True)
    last_retried_at = models.DateTimeField(null=True)

    objects = models.Manager()
    actual_objects = PushNotificationEventManager()

    class Meta:
        verbose_name = "push notification event"
        verbose_name_plural = "push notification events"
        unique_together = ["push_notification", "event"]
        indexes = [
            models.Index(fields=["processed_at", "retry", "last_retried_at"]),
        ]

    async def post(self, session: aiohttp.ClientSession) -> aiohttp.ClientResponse:
        headers = {
            "Content-Type": "application/json",
            "User-Agent": "SpotiiPush/1.0 (https://www.spotii.me/bot.html)",
        }

        push_notification = await PushNotification.actual_objects.aget(pk=self.push_notification_id)
        event = await Event.objects.aget(pk=self.event_id)

        async with session.post(
            push_notification.callback_url,
            json=event.data(),
            headers=headers
        ) as res:
            return res



import datetime
from pytz import utc
from rest_framework import mixins, viewsets, response, status
from push_notification.event import serializers
from push_notification.event import models


class PushNotificationViewSet(
        mixins.CreateModelMixin,
        mixins.ListModelMixin,
        mixins.RetrieveModelMixin,
        mixins.DestroyModelMixin,
        viewsets.GenericViewSet):
    """
    Push notifications is a convenient way to get notified about events from the system.
    One can register a callback, i.e a valid URL that will be called whenever there is an
    event dispatched for this subscriber. Note that this can result in a large numbers of calls
    """
    queryset = models.PushNotification.actual_objects.all().order_by('-created_at')
    serializer_class = serializers.PushNotificationSerializer
    authentication_classes = []
    permission_classes = []

    def create(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        push_notification, is_created = self.perform_create(serializer)

        return response.Response({
            'is_created': is_created,
            **self.get_serializer(instance=push_notification).data,
        }, status=status.HTTP_201_CREATED, headers=self.get_success_headers(serializer.data))

    def perform_create(self, serializer):
        # TODO: should we check not full path ?
        #  e.g. added earlier consumers/registration/created
        #  now he wants to add consumers/registration/*
        # push_notifications = models.PushNotification.objects.filter(
        #     callback_url=validated_data['callback_url'],
        # )
        # for push_notification in push_notifications:
        #     if fnmatch.fnmatch(push_notification.pattern, validated_data['pattern']):
        #         created = True
        #         return push_notification, created
        push_notification, created = models.PushNotification.objects.update_or_create(
            **serializer.data,
            defaults={
                'deleted_at': None
            }
        )
        return push_notification, created

    def perform_destroy(self, instance):
        instance.deleted_at = datetime.datetime.utcnow().replace(tzinfo=utc)
        instance.save()



from push_notification.event.models import Event


def emit_event(event_type, object_type, object_id, account_id=None, **kwargs):
    return Event.objects.create(
        event_type=event_type,
        object_type=object_type,
        object_id=object_id,
        account_id=account_id,
        **kwargs,
    )



# Generated by Mustafa on 2023-01-30 15:30

import django.core.serializers.json
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('event', '0002_auto_20200403_1401'),
    ]

    operations = [
        migrations.AlterField(
            model_name='event',
            name='meta_data',
            field=models.JSONField(blank=True, encoder=django.core.serializers.json.DjangoJSONEncoder, null=True),
        ),
    ]



# Generated by Django 2.2.5 on 2020-02-01 15:28

import django.contrib.postgres.fields.jsonb
from django.db import migrations, models
import django.db.models.deletion
import uuid


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Event',
            fields=[
                ('event_id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('event_type', models.CharField(max_length=254)),
                ('object_type', models.CharField(blank=True, max_length=254, null=True)),
                ('object_id', models.CharField(blank=True, max_length=255, null=True)),
                ('account_id', models.UUIDField(blank=True, null=True)),
                ('meta_data', django.contrib.postgres.fields.jsonb.JSONField(blank=True, null=True)),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('registerd_at', models.DateTimeField(null=True)),
            ],
            options={
                'verbose_name': 'event',
                'verbose_name_plural': 'events',
            },
        ),
        migrations.CreateModel(
            name='PushNotification',
            fields=[
                ('push_notification_id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('callback_url', models.URLField()),
                ('pattern', models.CharField(max_length=10000)),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('deleted_at', models.DateTimeField(null=True)),
            ],
            options={
                'verbose_name': 'push notification',
                'verbose_name_plural': 'push notifications',
                'unique_together': {('callback_url', 'pattern')},
            },
        ),
        migrations.CreateModel(
            name='PushNotificationEvent',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('retry', models.IntegerField(blank=True, null=True)),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('processed_at', models.DateTimeField(null=True)),
                ('last_retried_at', models.DateTimeField(null=True)),
                ('event', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='event.Event')),
                ('push_notification', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='event.PushNotification')),
            ],
            options={
                'verbose_name': 'push notification event',
                'verbose_name_plural': 'push notification events',
            },
        ),
        migrations.AddIndex(
            model_name='pushnotificationevent',
            index=models.Index(fields=['processed_at', 'retry', 'last_retried_at'], name='event_pushn_process_4728e8_idx'),
        ),
        migrations.AlterUniqueTogether(
            name='pushnotificationevent',
            unique_together={('push_notification', 'event')},
        ),
    ]



from push_notification.event.models import PushNotification
from rest_framework import serializers


class PushNotificationSerializer(serializers.ModelSerializer):
    class Meta:
        model = PushNotification
        fields = ['push_notification_id', 'callback_url', 'pattern', 'deleted_at']
        read_only_fields = ['push_notification_id', 'deleted_at']
        validators = [
        ]



from os import path

from setuptools import find_packages, setup

NAME = "spotii-push-notification2"
VERSION = "1.1.2"

REQUIRES = ["aiohttp", "django", "djangorestframework", "asgiref"]


this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name=NAME,
    version=VERSION,
    description="Spotii Push Notification",
    author_email="hello@nuclearo.com",
    url="",
    keywords=["Spotii", "Push Notification"],
    install_requires=REQUIRES,
    packages=find_packages(),
    include_package_data=True,
    long_description=long_description,
    long_description_content_type='text/markdown'
)



# Generated by Django 2.2.5 on 2020-04-03 14:01

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('event', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='pushnotification',
            name='callback_url',
            field=models.URLField(max_length=2000),
        ),
    ]
